import sys
import urllib3

from datetime import datetime
from urllib.parse import urlparse
from json.decoder import JSONDecodeError

from IzzyLib import logging
from IzzyLib.cache import LRUCache
from IzzyLib.misc import boolean
from DBUtils.PooledPg import PooledPg as DB
from tinydb import TinyDB, Query, where
from tinydb_smartcache import SmartCacheTable
from tinyrecord import transaction as trans
from tldextract import extract
from Crypto.PublicKey import RSA
from mastodon import Mastodon
from mastodon.Mastodon import MastodonUnauthorizedError, MastodonBadGatewayError, MastodonNetworkError

from .config import stor_path, PAWSCONFIG, MASTOCONFIG as mdb
from .functions import fetch, get_nodeinfo


def jsondb():
	try:
		db = TinyDB(f'{stor_path}/db.json', indent=4)

	except JSONDecodeError as e:
		logging.critical(f'Failed to load DB: {e}. Exiting...')
		sys.exit()

	db.table_class = SmartCacheTable

	class table:
		keys = db.table('keys')
		instances = db.table('instances')
		users = db.table('users')
		whitelist = db.table('whitelist')
		request = db.table('request')

	return table


def pgdb():
	try:
		if mdb['dbpass']:
			return DB(dbname=mdb['dbname'], host=mdb['dbhost'], port=mdb['dbport'], user=mdb['dbuser'], passwd=mdb['dbpass']).connection()

		else:
			return DB(dbname=mdb['dbname'], host=mdb['dbhost'], port=mdb['dbport'], user=mdb['dbuser']).connection()

	except Exception as e:
		logging.critical(f'Failed to connect to DB: {e}. Exiting...')
		sys.exit()


def keys(actor):
	if pawsdb.keys.get(query.actor == actor) == None:
		logging.info(f'No RSA key. Generating one for {actor}...')
		PRIV = RSA.generate(4096)
		PUB = PRIV.publickey()

		keydata = {
			'actor': actor,
			'pubkey': PUB.exportKey('PEM').decode('utf-8'),
			'privkey': PRIV.exportKey('PEM').decode('utf-8')
		}

		with trans(pawsdb.keys) as tr:
			tr.insert(keydata)


	return pawsdb.keys.get(query.actor == actor)


def get_handle(userid):
	user_data = mastodb.query(f'SELECT username,domain FROM public.accounts WHERE id = \'{userid}\'').dictresult()

	if len(user_data) < 1:
		return

	return (user_data[0]['username'].lower(), user_data[0]['domain'].lower())


def get_user(handle):
	user_data = mastodb.query(f'SELECT * FROM public.accounts WHERE username = \'{handle}\' and domain is NULL').dictresult()
	return user_data[0] if len(user_data) > 0 else None


def get_toot(tootid):
	toot = mastodb.query(f'SELECT * FROM public.statuses WHERE id = \'{tootid}\'').dictresult()
	return toot[0] if len(toot) > 0 else None


def get_bans(suspend=True, details=False):
	domains = mastodb.query('SELECT * FROM public.domain_blocks;').dictresult()
	banlist = {} if details else []

	for domain in domains:
		instance = domain['domain']
		severity = domain['severity']

		if suspend and severity != 1:
			continue

		if details:
			banlist[instance] = {
				'severity': domain['severity'],
				'media': boolean(domain['reject_media']),
				'reports': boolean(domain['reject_reports']),
				'private': domain['private_comment'],
				'public': domain['public_comment'],
				'updated': domain['updated_at']
			}

		else:
			banlist.append(instance)

	return banlist


def banned_user_check(access_user):
	if not access_user:
		return

	users = mastodb.query('SELECT username, domain FROM accounts WHERE suspended_at is not NULL').dictresult()

	if not users:
		return

	allbans = [(user['username'].lower(), user['domain'].lower()) for user in users]

	if access_user in allbans:
		return True


def ban_check(url):
	instance = urlparse(url).netloc if url.startswith('http') else url
	domain = extract(url)
	parsed = f'{domain.domain}.{domain.suffix}'
	banlist = get_bans()

	for ban in banlist:
		if parsed in ban or parsed == ban:
			return True

	logging.debug(f'{parsed} not in blocklist')


def user_ban_check(user, access_user):
	user_data = mastodb.query(f'SELECT id FROM public.accounts WHERE LOWER(username) = \'{user}\' and domain is NULL').dictresult()

	if len(user_data) < 1:
		return

	userid = user_data[0]['id']
	userban_query = mastodb.query(f'SELECT * FROM public.blocks WHERE account_id = \'{userid}\'').dictresult()
	userbans = []
	
	for ban in userban_query:
		acct_id = ban['target_account_id']
		user = get_handle(acct_id)

		if user:
			userbans.append(user)

		else:
			logging.warning(f'Invalid userid: {acct_id}')

	if access_user in userbans:
		return True

	return False


def user_domain_ban_check(handle, access_domain):
	user_data = get_user(handle)

	if not user_data:
		return

	userid = user_data['id']
	ban_data = mastodb.query(f'SELECT * FROM public.account_domain_blocks WHERE account_id = \'{userid}\' and domain = \'{access_domain}\'').dictresult()

	return ban_data


def wl_check(domain):
	data = pawsdb.whitelist.get(query.domain == domain)
	return data


def admin_check(handle):
	user = get_user(handle)

	if not user:
		return

	userid = user['id']
	user_data = mastodb.query(f'SELECT * FROM public.users WHERE account_id = \'{userid}\'').dictresult()

	if not user_data or not user_data[0]['admin']:
		return

	return True


def whitelist(action, instance):
	domain = pawsdb.whitelist.get(query.domain == instance)

	if action == 'add':
		if not domain:
			with trans(pawsdb.whitelist) as tr:
				tr.insert({'domain': instance})

	elif action == 'remove':
		if domain:
			with trans(pawsdb.whitelist) as tr:
				tr.remove(query.domain == instance)

	else:
		return 'InvalidAction'


def get_instances(domain=None, state=None):
	if domain:
		rows = pawsdb.instances.get(query.domain == domain)

	elif state:
		rows = pawsdb.instances.search(query.state == state)

	else:
		rows = pawsdb.instances.all()

	return rows


def instances(action, instance, state='request'):
	logging.debug(', '.join([action, instance, state]))

	if None in [action, instance, state]:
		return False

	domain = urlparse(instance).netloc if instance.startswith('http') else instance
	instance_data = pawsdb.instances.get(query.domain == domain)
	state = state.lower()

	if action == 'add':
		if instance_data:
			logging.debug(f'Updating instance state: {domain}, {state}')
			with trans(pawsdb.instances) as tr:
				tr.update({'state': state}, doc_ids=[instance_data.doc_id])

			return

		ni_software = get_nodeinfo(domain)

		if not ni_software:
			logging.debug(f'Failed to get nodeinfo data from instance: {domain}')
			return

		data = {
			'domain': domain,
			'software': ni_software['name'],
			'state': state,
			'timestamp': datetime.timestamp(datetime.now())
		}

		with trans(pawsdb.instances) as tr:
			tr.insert(data)

	elif action == 'remove':
		if not instance_data:
			logging.debug(f'Domain not in request list: {domain}')

		with trans(pawsdb.instances) as tr:
			tr.remove(doc_ids=[instance_data.doc_id])


def instance_check(domain):
	timestamp = datetime.timestamp(datetime.now())
	instance = pawsdb.instances.get(query.domain == domain)
	state = instance.get('state') if instance else None

	if not domain:
		return False

	if state == 'accept' or domain == mdb['domain']:
		return True

	elif state == 'deny':
		return False

	elif not state:
		result = instances('add', domain)

	return


def cleanup_users(invalid_check=None):
	timestamp = datetime.timestamp(datetime.now())
	invalid_offset = 60 * 15

	with trans(pawsdb.users) as tr:
		for user in pawsdb.users.all():
			if not user.get('timestamp'):
				logging.debug(f'adding timestamp for user: {user["handle"]}@{user["domain"]}')
				tr.update({'timestamp': timestamp}, doc_ids=[user.doc_id])
				continue

			if invalid_check:
				if user['timestamp'] < timestamp - invalid_offset and not user['token']:
					logging.debug(f'old and incomplete access token for {user["handle"]}@{user["domain"]}')
					tr.remove(doc_ids=[user.doc_id])
					continue

			else:
				if not user['token']:
					logging.debug(f'no access token for {user["handle"]}@{user["domain"]}')
					continue

				client = Mastodon(api_base_url=user['domain'], access_token=user['token'])

				try:
					client.me()
					logging.debug(f'valid token for {user["handle"]}@{user["domain"]}')

				except MastodonUnauthorizedError:
					logging.debug(f'invalid token for {user["handle"]}@{user["domain"]}')
					tr.remove(doc_ids=[user.doc_id])

				except (urllib3.exceptions.NewConnectionError, MastodonBadGatewayError, MastodonNetworkError):
					logging.debug(f'failed to connect to domain for {user["handle"]}@{user["domain"]}')


pawsdb = jsondb()
query = Query()
mastodb = pgdb()
