import json, socket, os, re
import aiohttp, validators, urllib3

from urllib.parse import urlparse

from IzzyLib import logging
from IzzyLib.template import aiohttpTemplate
from IzzyLib.cache import LRUCache

from .config import VERSION


httpclient = urllib3.PoolManager(num_pools=100, timeout=urllib3.Timeout(connect=15, read=15))


class cache:
	url = LRUCache()


def error(request, code, msg, isjson=False):
	if isjson or request.get('jsonreq'):
		return json_error(code, msg)

	else:
		return http_error(request, code, msg)


def json_error(code, error):
	error_body = json.dumps({'error': error})
	cont_type = 'application/json'
	return aiohttp.web.Response(body=error_body, content_type=cont_type, status=code)


def http_error(request, code, msg):
	data = {'msg': msg, 'code': str(code)}
	return aiohttpTemplate('error.html', data, request, status=code)


def fed_domain(user, domain):
	data = fetch(f'https://{domain}/.well-known/webfinger?resource=acct:{user}@{domain}')

	if not data:
		logging.error(f'User doesn\'t exist: {user}@{domain}')
		return

	wf_data = data.get('subject')
	user_data = wf_data.replace('acct:', '').split('@')

	return user_data[1]


def get_nodeinfo(instance):
	domain = urlparse(instance).netloc if instance.startswith('http') else instance

	nodeinfo = fetch(f'https://{domain}/nodeinfo/2.0.json')

	if not nodeinfo or (isinstance(nodeinfo, dict) and nodeinfo.get('error')):
		logging.debug('Wrong nodeinfo url. Finding correct one...')
		wk_url = f'https://{domain}/.well-known/nodeinfo'
		well_known = fetch(wk_url)

		if not well_known:
			logging.debug(f'failed to fetch {wk_url}')
			return

		ni_url = None

		try:
			for link in well_known['links']:
				if link['rel'].startswith('http://nodeinfo.diaspora.software/ns/schema/2'):
					ni_url = link['href']

		except Exception as e:
			logging.debug(f'Invalid nodeinfo data, Error: {e}')
			return

		nodeinfo = fetch(ni_url)

		if not nodeinfo:
			logging.debug(f'Failed to fetch nodeinfo for {domain}')

	return nodeinfo['software'] if nodeinfo else None


def domain_check(domain):
	stripped = domain.replace(' ', '')
	return stripped if validators.domain(stripped) else None


def user_check(path):
	parsed = path.replace('.json', '').split('/')

	if (parsed[1] == 'users' and len(parsed) < 4) or (parsed[1].startswith('@') and len(parsed) < 3):
		try:
			int(parsed[-1])
			return False

		except ValueError:
			return True

	return False


def parse_sig(signature, short=False):
	if not signature:
		return

	for line in signature.split(','):
		if 'keyid' in line.lower():
			actor = line.split('=')[1].split('#')[0].replace('"', '')
			return urlparse(actor).netloc if short else actor


def parse_ua(agent):
	if not agent:
		return

	url = re.findall('https?://(?:[-\w.]|(?:%[\da-fA-F]))+', agent)

	if not url:
		if 'mozilla' not in agent.lower():
			logging.debug('Failed to find url in user agent')

		return 'unknown'

	return urlparse(url[0]).netloc


def dig(domain):
	if not domain or domain == 'unknown':
		return

	try:
		return socket.gethostbyname(domain)

	except socket.gaierror as e:
		logging.info(f'Failed to resolve IP: {e}')


def distill_query(querydata):
	rawquery = ''

	if len(querydata) > 0:
		for var in querydata:
			if rawquery == '':
				rawquery += f'{var}={querydata[var]}'

			else:
				rawquery += f'&{var}={querydata[var]}'

	return rawquery if rawquery != '?' else ''


def css_ts():
	from .config import script_path

	css_check = lambda css_file : int(os.path.getmtime(f'{script_path}/templates/{css_file}.css'))

	color = css_check('color')
	layout = css_check('layout')

	return color + layout


def fetch(url, cached=True):
	if not url:
		logging.debug(f'Fetch was not fed a url')
		return

	cached_data = cache.url.fetch(url)

	if cached and cached_data:
		logging.debug(f'Returning cached data for {url}')
		return cached_data

	headers = {
		'User-Agent': f'PAWS/{VERSION}',
		'Accept': 'application/json'
	}

	try:
		logging.debug(f'Fetching new data for {url}')
		response = httpclient.request('GET', url, headers=headers)

	except Exception as e:
		logging.debug(f'Failed to fetch {url}')
		logging.debug(e)
		return

	if response.data == b'':
		logging.debug(f'Received blank data while fetching url: {url}')
		return

	try:
		data = json.loads(response.data)

		if cached:
			logging.debug(f'Caching {url}')
			cache.url.store(url, data)

		if data.get('error'):
			return

		return data

	except Exception as e:
		logging.debug(f'Failed to load data: {response.data}')
		logging.debug(e)
		return
