import asyncio, json, binascii, base64, traceback, aiohttp

from urllib.parse import urlparse, quote_plus, unquote_plus
from random import choice
from os.path import isfile
from http import client as HTTP

from IzzyLib import logging
from IzzyLib.misc import boolean
from aiohttp.http_exceptions import *
from aiohttp.client_exceptions import *
from Crypto.PublicKey import RSA

from .signature import validate, pass_hash, sign_headers
from .functions import error, user_check, domain_check, parse_sig, parse_ua, dig, distill_query, get_nodeinfo, httpclient
from .config import MASTOCONFIG, PAWSCONFIG, VERSION, script_path
from .database import pawsdb, query, trans, ban_check, user_ban_check, banned_user_check, wl_check, keys, user_domain_ban_check, admin_check, instance_check, instances


paws_host = PAWSCONFIG['domain']
masto_host = MASTOCONFIG['domain']
masto_path = PAWSCONFIG['mastopath']

blocked_agents = [
	'gabsocial',
	'kiwifarms',
	'fedichive',
	'liveview',
	'freespeech',
	'shitposter.club',
	'baraag',
	'gameliberty',
	'neckbeard',
	'soapbox',
	'qoto',
	'archive'
]

auth_paths = [
	'/@',
	'/users'
]

admin_paths = [
	'/paws/action',
	'/paws/list'
]

error_msgs = {
	404: 'Not found',
	500: 'Server did an oopsie'
}


def parse_headers(headers, enc=True):
	new_headers = {}

	for header in headers:
		key, value = [item.decode() for item in header]

		if key.lower() in ['content-encoding', 'transfer-encoding'] and not enc:
			continue

		if key in new_headers:
				new_headers[key] + '; ' + value

		else:
			new_headers[key] = value

	return new_headers


async def passthrough2(request, headers):
	mastohost = PAWSCONFIG['mastohost']
	raw_data = await request.read()
	is_json = request.get('jsonreq', False)
	Headers = parse_headers(request.raw_headers)
	Headers['Proxy'] = ''

	try:
		timeout = aiohttp.ClientTimeout(total=5)
		async with aiohttp.request(request.method, f'http://{mastohost}{request.path}?{request.query_string}', headers=Headers, data=raw_data, timeout=timeout, allow_redirects=False) as resp:
			data = await resp.read()

			if resp.status not in [200, 202, 301, 302]:
				if data in [b'Request not signed', 'Request not signed']:
					err_msg = 'Missing signature'
					logging.debug(err_msg)

				else:
					err_msg = f'Recieved error {resp.status} from Mastodon'
					logging.debug(err_msg)

				return error(request, resp.status, f'Failed to forward request')

			resp_headers = parse_headers(resp.raw_headers, False)

			return aiohttp.web.Response(body=data, headers=resp_headers, status=resp.status)

	except Exception as e:
		traceback.print_exc()
		return error(request, 504, f'Failed to connect to Mastodon')


async def passthrough(request, headers):
	mastohost = PAWSCONFIG['mastohost']
	req_data = await request.read() if request.body_exists else None
	is_json = request.get('jsonreq', False)

	try:
		resp = httpclient.request(request.method, f'http://{mastohost}{request.path}?{request.query_string}', body=req_data, headers=headers, redirect=False)
		data = resp.data

		if resp.status not in [200, 202, 301, 302]:
			if data in [b'Request not signed', 'Request not signed']:
				err_msg = 'Missing signature'
				logging.debug(err_msg)

			else:
				err_msg = f'Recieved error {resp.status} from Mastodon'
				logging.debug(err_msg)

			return error(request, resp.status, f'Failed to forward request')

		resp_headers = dict(resp.headers)

		if resp_headers.get('Content-Encoding'):
			del resp_headers['Content-Encoding']

		if resp_headers.get('Transfer-Encoding'):
			del resp_headers['Transfer-Encoding']

		return aiohttp.web.Response(body=data, headers=resp_headers, status=resp.status)

	except ClientConnectorError:
		traceback.print_exc()
		return error(request, 504, f'Failed to connect to Mastodon')


async def http_filter(app, handler):
	async def http_filter_handler(request):
		signature = request.headers.get('signature')
		sig_domain = parse_sig(signature, short=True)

		ua = request.headers.get('user-agent').lower()
		ua_domain = parse_ua(ua)
		domain = ua_domain if not sig_domain else sig_domain

		token = request.cookies.get('paws_token')
		user_data = None if not token else pawsdb.users.get(query.token == token)
		user = (user_data['handle'], user_data['instance']) if user_data and user_data.get('instance') else (None, None)

		real_ip = request.headers.get('X-Real-Ip', request.remote)
		ua_ip = dig(ua_domain)

		instance = domain if domain != 'unknown' else user[1]
		allow = instance_check(instance)
		request['jsonreq'] = True if 'json' in request.headers.get('Accept', '') or request.path.endswith('.json') else False

		# add logged in user data to the request for the frontend
		request['user'] = user_data
		request['admin'] = admin_check(user_data['handle']) if user_data else None

		# block nazis and general garbage
		for agent in blocked_agents:
			if agent in ua:
				logging.debug(f'Blocked garbage: {domain}')
				return error(request, 418, 'This teapot kills fascists')

		# Disable rss feeds
		if PAWSCONFIG['disable_rss'] and request.path.endswith('.rss'):
			return error(request, 403, 'RSS feeds disabled')

		if request.path in ['/paws/actor', '/paws/inbox', '/.well-known/webfinger'] and request.host != paws_host:
			return aiohttp.web.HTTPFound('/paws')

		# give up if a domain can't be found'
		if not domain:
			return error(request, 401, 'Can\'t find instance domain')

		# block any suspended instances
		if ban_check(domain):
			logging.debug(f'Blocked instance: {domain}')
			return error(request, 403, 'Forbidden')

		# block any suspended users
		if banned_user_check(user):
			logging.debug(f'Blocked user: {domain}')
			return error(request, 403, 'Access Denied')

		# prevent unauthed users from accessing the instance lists
		if any(map(request.path.startswith, admin_paths)) and not request['admin']:
			return aiohttp.web.HTTPFound('/paws/login')

		if any(map(request.path.startswith, auth_paths)) and request.method == 'GET':
			# Check signatures if auth fetches are off
			if not user_check(request.path):
				if PAWSCONFIG['require_approval'] and not allow:
					if allow != False:
						status, message = (401, 'Instance awaiting approval or rejection')
						instances('add', instance)

					else:
						status, message = (403, 'Rejected')

					return error(request, status, message)

				if not MASTOCONFIG['auth_fetch']:
					if signature:
						actor = parse_sig(signature)

						if not (await validate(actor, request)):
							logging.warning(f'Signature validation failed for: {actor}')
							return error(request, 401, 'signature check failed, signature did not match key')

					elif real_ip == ua_ip and wl_check(domain):
						logging.info(f'Letting {domain} through')

					else:
						msg = 'missing signature'
						logging.warning(msg)
						return error(request, 401, msg)

			if not request['jsonreq']:
				if not token or not user_data:
					return aiohttp.web.HTTPFound(f'/paws/login?redir={quote_plus(request.path)}')

				split_path = request.path.split('/')
				user = split_path[1].replace('@', '') if request.path.startswith('/@') else split_path[2]

				if user_ban_check(user.lower(), (user_data['handle'].lower(), user_data['instance'])) or user_domain_ban_check(user.lower(), user_data['instance']):
					return error(request, 403, 'Access Denied')

		if signature and wl_check(domain) and request.method == 'GET':
			logging.warning(f'{domain} has started signing requests and can be removed from the whitelist')

		if not signature and real_ip == ua_ip and wl_check(domain) and request.method == 'GET':
			logging.debug(f'Signing fetch for whitelisted instance: {domain}')

			HEADERS = {
				'(request-target)': f'get {request.path}',
				'Accept': 'application/json',
				'User-Agent': f'PAWS/{VERSION}; https://{masto_host}',
				'Host': masto_host,
				'X-Forwarded-For': request.headers.get('X-Forwarded-For', request.remote),
				'X-Forwarded-Port': '443',
				'X-Forwarded-Proto': 'https',
				'X-Real-Ip': request.headers.get('X-Real-Ip', request.remote)
			}
			HEADERS['signature'] = sign_headers(HEADERS, 'default', f'https://{paws_host}/paws/actor#main-key')
			HEADERS.pop('(request-target)')

		else:
			HEADERS = request.headers

		if not request.path.startswith('/paws'):
			masto_file = f'{masto_path}/public{request.path}'

			if isfile(masto_file):
				return aiohttp.web.FileResponse(masto_file, headers=HEADERS)

			return_data = await passthrough(request, HEADERS)
			return return_data

		return (await handler(request))
	return http_filter_handler


# Fucking trailing slashes
async def http_trailing_slash(app, handler):
	async def http_trailing_slash_handler(request):
		if request.path != '/' and request.path.endswith('/'):
			return aiohttp.web.HTTPFound(request.path[:-1])
		
		return (await handler(request))
	return http_trailing_slash_handler


# Custom error handler
async def http_error(app, handler):
	async def http_error_handler(request):
		try:
			response = await handler(request)

		#except aiohttp.web.HTTPException as ex:
		#	message = error_msgs.get(ex.status, 'Server did an oopsie')
		#	response = error(request, ex.status, message)

		except Exception as e:
			if getattr(e, 'status', None):
				message = error_msgs.get(e.status, 'Server did an oopsie')
				response = error(request, e.status, message)

			else:
				traceback.print_exc()
				response = error(request, 500, e)

		return response
	return http_error_handler


async def http_server_header(request, response):
	response.headers['Server'] = f'PAWS/{VERSION}'

	phrases = [
		'trans rights are human rights',
		'trans rights',
		'be gay, do crimes',
		'punch nazis',
		'eat the rich',
		'im gay',
		'heck'
	]

	value = choice(phrases)

	response.headers['msg'] = value


async def http_access_log(request, response):
	if not request.path.endswith(('js', 'css', 'png', 'jpg', 'gif')):
		uagent = request.headers.get('user-agent')
		client_ip = request.headers.get('X-Real-IP', request.remote)

		logging.info(f'{client_ip} {request.method} {request.path_qs} {response.status} "{uagent}"')
