import os, sys, asyncio, aiohttp, aiohttp_jinja2, jinja2

from IzzyLib import logging, template, color
from IzzyLib.template import buildTemplates, templateWatcher
from ipaddress import ip_address as address
from urllib.parse import urlparse
from mastodon import Mastodon

from .config import PAWSCONFIG, MASTOCONFIG, VERSION, script_path
from .functions import css_ts
from .database import cleanup_users, admin_check
from . import middleware, views, stor_path


def webserver():
	web = aiohttp.web.Application(middlewares=[
		middleware.http_filter,
		middleware.http_trailing_slash,
		middleware.http_error
	])

	web.on_response_prepare.append(middleware.http_server_header)
	web.on_response_prepare.append(middleware.http_access_log)

	web.add_routes([
		aiohttp.web.route('GET', '/paws', views.get_paws),
		aiohttp.web.route('GET', '/paws/imgay', views.get_gay),
		aiohttp.web.route('POST', '/paws/action/{action}', views.post_paws),
		aiohttp.web.view('/paws/list/{list}', views.lists),
		aiohttp.web.view('/paws/login', views.login),
		aiohttp.web.route('GET', '/paws/logout', views.get_logout),
		aiohttp.web.route('GET', '/paws/auth', views.get_auth),
		aiohttp.web.route('GET', '/paws/style-{timestamp}.css', views.get_style),

		aiohttp.web.route('GET', '/.well-known/webfinger', views.get_webfinger),
		aiohttp.web.route('POST', '/paws/inbox', views.post_inbox),
		aiohttp.web.route('GET', '/paws/actor', views.get_actor),

		#aiohttp.web.route('GET', '/paws/outbox', views.get_outbox),
		#aiohttp.web.route('GET', '/paws/imgay', views.get_gay)
	])

	template.addEnv({
		'css_ts': css_ts,
		'domain': MASTOCONFIG['domain'],
		'VERSION': VERSION,
		'len': len,
		'urlparse': urlparse,
		'admin_check': admin_check
	})

	return web


async def start_webserver():
	app = webserver()
	runner = aiohttp.web.AppRunner(app, access_log_format='%{X-Real-Ip}i %s %b "%r" "%{User-Agent}i"')
	await runner.setup()

	listen = PAWSCONFIG['host']
	port = PAWSCONFIG['port']

	if listen.startswith('unix:'):
		if sys.platform != 'win32':
			sock_listen = listen.replace('unix:', '')

			logging.info(f'Starting webserver at socket: {sock_listen}')
			site = aiohttp.web.UnixSite(runner, sock_listen)
			os.chmod(sock_listen, 0o664)

		else:
			logging.error('Windows cannot use unix sockets. Use an IP address instead. Exiting...')
			sys.exit()

	else:
		try:
			address(listen)

		except ValueError:
			logging.warning('Invalid IP address. Listening on "127.0.0.1" instead.')
			listen = '127.0.0.1'

		try:
			int(port)

		except ValueError:
			logging.warning('Invalid port. Using 3001 instead.')
			port = 3001

		logging.info(f'Starting webserver at address: {listen}:{port}')
		site = aiohttp.web.TCPSite(runner, listen, port)

	await site.start()


async def cleanup_tokens():
	while True:
		await asyncio.sleep(60*60)
		logging.debug('Cleaning up tokens')
		cleanup_users()


async def cleanup_invalid_users():
	while True:
		logging.debug('Cleaning up invalid users')
		cleanup_users(True)
		await asyncio.sleep(60*15)


def main():
	buildTemplates()
	observer = templateWatcher()

	try:
		loop = asyncio.get_event_loop()
		asyncio.ensure_future(start_webserver())
		asyncio.ensure_future(cleanup_tokens())
		asyncio.ensure_future(cleanup_invalid_users())
		loop.run_forever()

	except KeyboardInterrupt:
		logging.info('Bye!')

	observer.stop()

