import aiohttp
import random
import traceback

from operator import itemgetter

from IzzyLib import logging
from IzzyLib.cache import TTLCache
from IzzyLib.template import aiohttpTemplate
from urllib.parse import quote_plus, unquote_plus, urlparse
from datetime import datetime

from . import oauth
from .database import pawsdb, trans, query, where, keys, ban_check, get_user, get_toot, admin_check, whitelist, instances
from .functions import error, fed_domain, domain_check, css_ts
from .config import MASTOCONFIG, PAWSCONFIG


paws_host = PAWSCONFIG['domain']
masto_host = MASTOCONFIG['domain']


class login(aiohttp.web.View):
	async def get(self):
		parms = self.request.rel_url.query
		redir = parms.get('redir')
		numid = random.randint(1*1000000, 10*1000000-1)

		if self.request['user']:
			return aiohttp.web.HTTPFound('/paws')

		data = {'redir': redir, 'numid': numid}
		return aiohttpTemplate('login.html', data, self.request)


	async def post(self):
		data = await self.request.post()
		domain = data.get('domain')
		redir = data.get('redir')
		numid = data.get('numid')

		if 'DROP ' in domain:
			return aiohttp.web.HTTPFound('https://www.youtube.com/watch?v=dQw4w9WgXcQ')

		domain = domain_check(domain)

		if not domain:
			return error(self.request, 400, 'Invalid domain')

		if ban_check(domain):
			return error(self.request, 403, 'Instance banned')

		appid, appsecret, redir_url = oauth.create_app(domain)

		if appid == 'error':
			return error(self.request, 500, appsecret)


		with trans(pawsdb.users) as tr:
			tr.insert({
				'handle': data['numid'],
				'domain': data['domain'].lower(),
				'appid': appid,
				'appsecret': appsecret,
				'token': None,
				'timestamp': datetime.timestamp(datetime.now())
			})

		response = aiohttp.web.HTTPFound(redir_url)
		response.set_cookie('paws_numid', numid, max_age=60*60, path='/paws')

		if redir not in ['', None]:
			response.set_cookie('paws_redir', redir, max_age=60*60, path='/paws')

		return response


class lists(aiohttp.web.View):
	async def get(self):
		request = self.request
		listtype = request.match_info['list']
		state = request.query.get('type', None)

		if request['admin']:
			if state not in ['accept', 'deny']:
				instances = pawsdb.instances.search(query.state != 'request')

			else:
				instances = pawsdb.instances.search(query.state == state)

			requests = pawsdb.instances.search(query.state == 'request')
			signlist = [line['domain'] for line in pawsdb.whitelist.all()]
			signlist.sort()
			data = {
				'listtype': listtype,
				'signlist': signlist,
				'requests': sorted(requests, key=lambda k: k['domain']),
				'instances': sorted(instances, key=lambda k: k['domain'])
			}

		else:
			data = {}

		if listtype not in data.keys():
			return error(request, 404, 'Invalid list type')

		return aiohttpTemplate('lists.html', data, request)

	async def post(self):
		pass


async def get_home(request):
	return aiohttpTemplate('home.html', {}, request)


async def get_paws(request):
	data = {}
	return aiohttpTemplate('panel.html', data, request)


async def post_paws(request):
	data = await request.post()
	user_data = request['user']
	domain = data.get('name')
	action = request.match_info['action'].lower()
	action = 'add' if action == 'update' else action
	admin = admin_check(user_data['handle']) if user_data else None
	page = request.query.get('list', 'instances')

	if not admin:
		return error(request, 403, 'Not an admin')

	if None in [action, domain]:
		return error(request, 400, 'Missing action or doamin')

	domain = urlparse(domain.replace(' ', ''))
	parsed_domain = domain.netloc if domain.netloc != '' else domain.path

	if action not in ['add', 'remove']:
		error(request, 400, 'Invalid action')

	result = instances(action, parsed_domain, data.get('action', 'request'))
	print(result)

	return aiohttp.web.HTTPFound(f'/paws/list/{page}')


async def get_auth(request):
	parms = request.rel_url.query
	cookies = request.cookies
	redir = cookies.get('paws_redir')
	numid = cookies.get('paws_numid')
	code = parms.get('code')

	if None in [numid, code]:
		return error(request, 500, 'Missing temporary userid or auth code')

	if redir in ['', 'None', None]:
		redir = '/paws'

	user = pawsdb.users.get(query.handle == str(numid))
	token, userinfo = oauth.login(user, code)

	if token == 'error':
		return error(request, 500, userinfo)

	instance = fed_domain(userinfo['username'], user['domain'])

	with trans(pawsdb.users) as tr:
		tr.update(
			{
				'handle': userinfo['username'],
				'instance': instance,
				'appid': None,
				'appsecret': None,
				'token': token
			},
			where('handle') == numid
		)

	response = aiohttp.web.HTTPFound(redir)
	response.set_cookie('paws_token', token, max_age=60*60*24*14)
	response.del_cookie('paws_redir', path='/paws')
	response.del_cookie('paws_numid', path='/paws')
	return response


async def get_logout(request):
	if not request['user']:
		return aiohttp.web.HTTPFound('/paws/login')

	token_id = request['user'].doc_id

	with trans(pawsdb.users) as tr:
		tr.remove(doc_ids=[token_id])

	response = aiohttp.web.HTTPFound('/paws/login')
	response.del_cookie('token')

	return response


async def get_style(request):
	maxage = 60*60*24*7

	response = aiohttpTemplate('color.css', {}, request, content_type='text/css')
	response.headers['Cache-Control'] = 'public,max-age={maxage},immutable'

	return response


async def get_webfinger(request):
	res = request.rel_url.query.get('resource')

	if not res or res != f'acct:paws@{paws_host}':
		data = {}

	else:
		data = {
			"aliases": [
				f"https://{paws_host}/paws/actor"
			],
			"links": [
				{
					"href": f"https://{paws_host}/paws/actor",
					"rel": "self",
					"type": "application/activity+json"
				},
				{
					"href": f"https://{paws_host}/paws/actor",
					"rel": "self",
					"type": "application/ld+json; profile=\"https://www.w3.org/ns/activitystreams\""
				}
			],
			"subject": f"acct:paws@{paws_host}"
		}

	return aiohttp.web.json_response(data)


async def get_actor(request):
	rsakey = keys('default')

	if not rsakey:
		return error(request, 500, 'Missing actor keys')

	PUBKEY = rsakey['pubkey']

	data = {
		'@context': [
			'https://www.w3.org/ns/activitystreams'
		],
		#'endpoints': {
			#'sharedInbox': f'https://{masto_host}/inbox'
		#},
		#'following': f'https://{host}/paws/following',
		#'followers': f'https://{host}/paws/followers',
		'inbox': f'https://{paws_host}/paws/inbox',
		#'outbox': f'https://{host}/paws/outbox',
		'name': f'PAWS @ {masto_host}',
		'type': 'Application',
		'id': f'https://{paws_host}/paws/actor',
		'publicKey': {
			'id': f'https://{paws_host}/paws/actor#main-key',
			'owner': f'https://{paws_host}/paws/actor',
			'publicKeyPem': PUBKEY
		},
		'summary': 'PAWS Actor',
		'preferredUsername': 'paws',
		'url': f'https://{paws_host}/paws/actor'
	}

	return aiohttp.web.json_response(data)


async def post_inbox(request):
	return aiohttp.web.json_response('UvU', status=202)


# failed attempt at creating a status
async def get_outbox(request):
	if not request.rel_url.query.get('page'):
		data = {
			"@context": "https://www.w3.org/ns/activitystreams",
			"id": f"https://{paws_host}/paws/outbox",
			"type": "OrderedCollection",
			"totalItems": 1,
			"first": f"https://{paws_host}/paws/outbox?page=true",
			"last": f"https://{paws_host}/paws/outbox?page=true"
	}

	else:
		data = {
			"@context": [
				"https://www.w3.org/ns/activitystreams",
				{
					"ostatus": "http://ostatus.org#",
					"atomUri": "ostatus:atomUri",
					"inReplyToAtomUri": "ostatus:inReplyToAtomUri",
					"conversation": "ostatus:conversation",
					"sensitive": "as:sensitive",
					"toot": "http://joinmastodon.org/ns#",
					"votersCount": "toot:votersCount",
					"blurhash": "toot:blurhash",
					"focalPoint": {
						"@container": "@list",
						"@id": "toot:focalPoint"
					},
					"Emoji": "toot:Emoji"
				}
			],
			"id": f"https://{paws_host}/paws/outbox?page=true",
			"type": "OrderedCollectionPage",
			"partOf": f"https://{paws_host}/paws/outbox",
			"orderedItems": [imgay]
		}

	return aiohttp.web.json_response(data)

async def get_gay(request):
	data = imgay

	return aiohttp.web.json_response(data)


imgay = {
	"id": f"https://{paws_host}/paws/imgay",
	"type": "Note",
	"published": '2020-01-20T09:45:00Z',
	"attributedTo": f"https://{paws_host}/paws/actor",
	"content": '<p>im gay</p>',
	"to": [
		"https://www.w3.org/ns/activitystreams#Public"
	],
	"cc": [
		f"https://{paws_host}/paws/actor/followers"
	]
}
